{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Description:  Wave audio direct-to-disk recorder using low-level multimedia
              wave audio API and real-time audio compression.
              This program shows how to perform direct-to-disk recording and
              playing of wave audio using a dual-buffer procedure.  While one
              buffer is being filled with recorded data, the other is being
              written to disk. RecDemo also includes function calls to the
              Audio Compression Manager (ACM) for setting the wave audio format.
              This allows various compression formats to be selected by the
              user.  When compresison is used, the size of the wave file is
              reduced significantly while retaining high audio quality.
              This program is based on DDREC Windows SDK sample program.
Author:       (c) 1999, Franois Piette. All rights reserved.
              http://www.rtfm.be/fpiette/indexuk.htm
              francois.piette@rtfm.be
              francois.piette@pophost.eunet.be
Creation:     June 16, 1999
Version       1.00
Legal Stuff:  This software may be used without charge provided this notice is
              not removed. An acknowledgement for the author's work is welcome.

              THIS SOFTWARE IS PROVIDED BY FRANCOIS PIETTE "AS IS" AND ANY
              EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
              THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
              PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR
              OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
              SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
              LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
              USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
              AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
              LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
              ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
              POSSIBILITY OF SUCH DAMAGE.
Support:      This code is unsupported. Anyway, I'll do my best to make it
              works correctly. If you find any problem, send me an EMail, I'll
              see if I can do something for you.
History:
Jul 21, 1999  V1.00 Released


 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
unit RecDemo1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  IniFiles, StdCtrls, ExtCtrls, MsAcm, MMSystem, Menus;

type
  TRecorderMode     = (recModeOff, recModeRecord, recModePlay);
  TRecorderSaveMode = (smSave, smSaveAs);
  TRecorderForm = class(TForm)
    ToolPanel: TPanel;
    DisplayMemo: TMemo;
    PlayButton: TButton;
    AudioFormatButton: TButton;
    Panel1: TPanel;
    Label1: TLabel;
    LengthDispLabel: TLabel;
    Label2: TLabel;
    LengthPosLabel: TLabel;
    FormatTagLabel: TLabel;
    FormatDescLabel: TLabel;
    RecordButton: TButton;
    PositionPanel: TPanel;
    Timer1: TTimer;
    RecordLightShape: TShape;
    OpenDialog1: TOpenDialog;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Open1: TMenuItem;
    Save1: TMenuItem;
    SaveAs1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    Help1: TMenuItem;
    PositionShape: TShape;
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure AudioFormatButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure RecordButtonClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure Save1Click(Sender: TObject);
    procedure SaveAs1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure PlayButtonClick(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure Help1Click(Sender: TObject);
  private
    FIniFileName   : String;
    FInitialized   : Boolean;
    FWaveFormat    : PWAVEFORMATEX;
    FTotalWaveSize : DWORD;                    // Number of samples recorded
    FByteDataSize  : DWORD;                    // Accumulative size of recorded data
    FDiskFreeSpace : DWORD;                    // Free space for temp file
    FWaveHdr       : array [0..1] of PWAVEHDR; // Points to wave header information
    FWaveMem       : array [0..1] of PChar;    // Points to wave buffers
    FBufIndex      : Integer;                  // Which buffer are we using
    FWaveIn        : HWAVEIN;                  // Device ID for recording input
    FWaveOut       : HWAVEOUT;                 // Device ID for playing
    FMaxFmtSize    : DWORD;                   // largest format size required for compression
    FFormatDesc    : String;                   // Format description
    FFormatTag     : String;                   // Format tag description
    FDeviceOpened  : Boolean;                  // Device open status
    FRecorderMode  : TRecorderMode;            // Recording/playing mode
    FWaveBufSize   : DWORD;                    // Size of buffer
    FFilename      : String;                   // File name to store WAV
    FHasFilename   : Boolean;
    FFileSaved     : Boolean;                  // Set when a save is done
    FMoreToPlay    : Boolean;                  // There is more data to play
    FRecordedData  : Boolean;                  // Did we recorded data
    FTmpFileName   : String;                   // temporary wave filename
    FTmpFileHandle : HFILE;                    // Handle to temporary wave file
    procedure Display(Msg : String);
    function  GetWaveFormat : integer;
    procedure UpdateLength(BytePosition : DWORD; BytePositiontotal : DWORD);
    function  GetFormatDetails(pfmtin : PWAVEFORMATEX) : integer;
    function  GetFormatTagDetails(wFormatTag : WORD) : integer;
    procedure errormsg(const msg : String);
    procedure ierrormsg(const msg : String);
    procedure WFerror(mmfp : HMMIO; const msg : String);
    procedure DestroyWaveRecorder;
    function  InitWaveRecorder : integer;
    function  AllocWaveFormatEx : Integer;
    procedure FreeWaveFormatEx;
    function  AllocWaveHeader : integer;
    procedure FreeWaveHeader;
    function  AllocPCMBuffers : Integer;
    procedure FreePCMBuffers;
    function  OpenTmpFile : Integer;
    function  CreateTmpFile : integer;
    procedure DeleteTmpFile;
    procedure CloseTmpFile;
    function  StartWaveRecord : Integer;
    procedure StopWaveRecord;
    procedure SetButtonState;
    procedure InitWaveHeaders;
    procedure CloseWaveDeviceRecord;
    procedure UpdateRecordDisplay;
    procedure UpdatePlayDisplay;
    function  AddNextBuffer : integer;
    procedure RecordLight(Mode : TRecorderMode);
    procedure MMWimData(var msg: TMessage); message MM_WIM_DATA;
    procedure MMWomDone(var msg: TMessage); message MM_WOM_DONE;
    procedure MMWomClose(var msg: TMessage); message MM_WOM_CLOSE;
    function  WriteWaveBuffer(size : UINT) : integer;
    procedure CheckSaveFile;
    function  SaveWaveFile(FSaveMode : TRecorderSaveMode) : integer;
    procedure InitFilenameDialog(OpenDialog : TOpenDialog);
    function  QueryUserSave(const name : String) : integer;
    function  GetSaveFilename(FSaveMode : TRecorderSaveMode) : Boolean;
    function  CopyDataToWaveFile(mmfp : HMMIO) : integer;
    function  StartWavePlay : Integer;
    function  QueueNextBuffer : Integer;
    function  ReadWaveBuffer : Integer;
    procedure StopWavePlay(prematurely : Boolean);
    procedure CloseWaveDevicePlay;
    function  ReadWaveFile : Integer;
    function  CopyWaveToTempFile(mmfp : HMMIO; datasize : DWORD) : Integer;
  public
    property IniFileName : String read FIniFileName write FIniFileName;
  end;

const
  WAVE_BUFSIZE  = 32768;
  FOURCC_WAVE   = $45564157;   { 'WAVE' }
  FOURCC_FMT    = $20746d66;   { 'fmt ' }
  FOURCC_FACT   = $74636166;   { 'fact' }
  FOURCC_DATA   = $61746164;   { 'data' }
  WindowCaption = 'Dual-buffer recorder';

var
  RecorderForm: TRecorderForm;

implementation

uses RecDemo2;

{$R *.DFM}

const
    SectionWindow      = 'RecDemoWindow';
    KeyTop             = 'Top';
    KeyLeft            = 'Left';
    KeyWidth           = 'Width';
    KeyHeight          = 'Height';


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.FormCreate(Sender: TObject);
begin
    FIniFileName := LowerCase(ExtractFileName(Application.ExeName));
    FIniFileName := Copy(FIniFileName, 1, Length(FIniFileName) - 3) + 'ini';
    FRecorderMode  := recModeOff;
    FFilename      := 'Untitled.wav';
    if InitWaveRecorder <> 0 then
        Application.Terminate;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.FormDestroy(Sender: TObject);
begin
    if FRecorderMode = recModeRecord then
        StopWaveRecord
    else if FRecorderMode = recModePlay then
        StopWavePlay(TRUE);
    DestroyWaveRecorder;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.FormShow(Sender: TObject);
var
    IniFile : TIniFile;
begin
    if not FInitialized then begin
        FInitialized := TRUE;

        IniFile      := TIniFile.Create(FIniFileName);
        Width        := IniFile.ReadInteger(SectionWindow, KeyWidth,  Width);
        Height       := IniFile.ReadInteger(SectionWindow, KeyHeight, Height);
        Top          := IniFile.ReadInteger(SectionWindow, KeyTop,
                                            (Screen.Height - Height) div 2);
        Left         := IniFile.ReadInteger(SectionWindow, KeyLeft,
                                            (Screen.Width  - Width)  div 2);
        IniFile.Destroy;

        Display('');
        DisplayMemo.Clear;
        DisplayMemo.Visible := FALSE;
        ToolPanel.Align     := alClient;

        Caption                 := WindowCaption + ' - ' + FFileName;
        LengthPosLabel.Caption  := '0';
        LengthDispLabel.Caption := '0';
        FormatTagLabel.Caption  := FFormatTag;
        FormatDescLabel.Caption := FFormatDesc;
        RecordLight(recModeOff);
        UpdateLength(0, 0);
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.FormClose(Sender: TObject; var Action: TCloseAction);
var
    IniFile : TIniFile;
begin
    IniFile := TIniFile.Create(FIniFileName);
    IniFile.WriteInteger(SectionWindow, KeyTop,         Top);
    IniFile.WriteInteger(SectionWindow, KeyLeft,        Left);
    IniFile.WriteInteger(SectionWindow, KeyWidth,       Width);
    IniFile.WriteInteger(SectionWindow, KeyHeight,      Height);
    IniFile.Destroy;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.Display(Msg : String);
var
    I : Integer;
begin
    DisplayMemo.Lines.BeginUpdate;
    try
        if DisplayMemo.Lines.Count > 200 then begin
            for I := 1 to 50 do
                DisplayMemo.Lines.Delete(0);
        end;
        DisplayMemo.Lines.Add(Msg);
    finally
        DisplayMemo.Lines.EndUpdate;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Display an error message.
procedure TRecorderForm.errormsg(const msg : String);
begin
    Application.MessageBox(PChar(msg), 'Error', MB_OK);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Display an error message and return an error value.
procedure TRecorderForm.ierrormsg(const msg : String);
begin
    Application.MessageBox(PChar(msg), 'Error', MB_OK);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Get the format tag details and store the string description.
function TRecorderForm.GetFormatTagDetails(wFormatTag : WORD) : integer;
var
    acmtagdetails : TACMFORMATTAGDETAILS;
begin
    // zero out...
    FillChar(acmtagdetails, sizeof(acmtagdetails), 0);

    acmtagdetails.cbStruct    := sizeof(acmtagdetails);
    acmtagdetails.dwFormatTag := wFormatTag;

    if acmFormatTagDetails(nil, acmtagdetails,
                           ACM_FORMATTAGDETAILSF_FORMATTAG) <> 0 then begin
        ierrormsg('Warning, FormatTagDetails function failed');
        Result := -1;
        Exit;
    end;

    // store the format tag details description string...
    FFormatTag := acmtagdetails.szFormatTag;
    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Get the format details and store the string description.
function TRecorderForm.GetFormatDetails(pfmtin : PWAVEFORMATEX) : integer;
var
    acmfmtdetails : TACMFORMATDETAILS;
begin
    // zero out struct...
    FillChar(acmfmtdetails, sizeof(acmfmtdetails), 0);

    acmfmtdetails.cbStruct    := sizeof(acmfmtdetails);
    acmfmtdetails.pwfx        := pfmtin;
    acmfmtdetails.dwFormatTag := pfmtin.wFormatTag;
    acmfmtdetails.cbwfx       := sizeof(TWAVEFORMATEX) + pfmtin.cbSize;

    if acmFormatDetails(nil, acmfmtdetails,
                        ACM_FORMATDETAILSF_FORMAT) <> 0 then begin
        ierrormsg('Warning, FormatDetails function failed');
        Result := -1;
        Exit;
    end;

    // store the format details description string...
    FFormatDesc := acmfmtdetails.szFormat;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Update the display of the sample length.
procedure TRecorderForm.UpdateLength(BytePosition : DWORD; BytePositiontotal : DWORD);
begin
    // store these...
    LengthPosLabel.Caption  := IntToStr(BytePosition);
    LengthDispLabel.Caption := IntToStr(BytePositiontotal);
    if (BytePosition = 0) or (BytePositiontotal = 0) then
       PositionShape.Width := 0
    else
       PositionShape.Width := Trunc((PositionPanel.Width - 2) *
                                     Integer(BytePosition) / Integer(BytePositiontotal));
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
function TRecorderForm.GetWaveFormat : integer;
var
    acmopt  : TACMFORMATCHOOSE;
    err     : MMRESULT;
    ptmpfmt : PWAVEFORMATEX;
begin
    // store the format temporarily...
    GetMem(ptmpfmt, FMaxFmtSize);
    if ptmpfmt = nil then begin
        ierrormsg('Error allocating temporary format buffer.');
        Result := -1;
        Exit;
    end;

    Move(FWaveFormat^, ptmpfmt^, FMaxFmtSize);

    // setup ACM choose fields and display the dialog...
    FillChar(acmopt, sizeof(acmopt), 0);  // zero out
    acmopt.cbStruct  := sizeof(acmopt);
    acmopt.fdwStyle  := ACMFORMATCHOOSE_STYLEF_INITTOWFXSTRUCT;
    acmopt.hwndOwner := Handle;
    acmopt.pwfx      := FWaveFormat;
    acmopt.cbwfx     := FMaxFmtSize;
    acmopt.pszTitle  := 'Select Compression';
    acmopt.fdwEnum   := ACM_FORMATENUMF_INPUT;
    err              := acmFormatChoose(acmopt);

    // if the same format was selected we don't want to reset FTotalWaveSize
    // below, so act like a cancel...
    if CompareMem(FWaveFormat, ptmpfmt, sizeof(TWAVEFORMATEX)) then
        err := ACMERR_CANCELED;
    if err <> MMSYSERR_NOERROR then begin
        Move(ptmpfmt^, FWaveFormat^, FMaxFmtSize);
        FreeMem(ptmpfmt);
        if err = ACMERR_CANCELED then begin
            Result := 0;
            Exit;
        end;
        ierrormsg('Error in FormatChoose function');
        Result :=  -2;
        Exit;
    end;

    // store the format description...
    FFormatDesc := acmopt.szFormat;

    // get the format tag details, we don't need to call acmGetFormatDetails since
    // that information was supplied by the choose function...
    GetFormatTagDetails(acmopt.pwfx.wFormatTag);

    FreeMem(ptmpfmt);

    // now set the play button to a grayed state cause we don't want
    // to try to play the recorded data with a different format...
    FTotalWaveSize := 0;
    UpdateLength(FTotalWaveSize, 0);
    SetButtonState;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.AudioFormatButtonClick(Sender: TObject);
begin
    GetWaveFormat;
    FormatTagLabel.Caption  := FFormatTag;
    FormatDescLabel.Caption := FFormatDesc;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Allocate format and wave headers, data buffers, and temporary filename.
function TRecorderForm.InitWaveRecorder : integer;
var
    Temp : array [0..MAX_PATH] of char;
begin
    Result := -1;
    // allocate memory for wave format structure...
    if AllocWaveFormatEx <> 0 then
        Exit;

    // find a device compatible with the available wave characteristics...
    if waveInGetNumDevs < 1 then begin
        ierrormsg('No wave audio recording devices found.');
        Result := -1;
        Exit;
    end;

    // allocate the wave header memory...
    if AllocWaveHeader <> 0 then begin
        Result := -3;
        Exit;
    end;

    // allocate the wave data buffer memory...
    if AllocPCMBuffers <> 0 then begin
        Result := -4;
        Exit;
    end;

    // generate a temporary filename for writing to...
    GetTempPath(sizeof(Temp), Temp);
    SetLength(FTmpFileName, MAX_PATH);
    GetTempFileName(Temp, 'wr', 0, PChar(FTmpFileName));

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Free the memory associated with the wave buffers.
procedure TRecorderForm.DestroyWaveRecorder;
begin
    FreeWaveFormatEx;
    FreePCMBuffers;
    FreeWaveHeader;
    DeleteTmpFile;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Allocate and lock WAVEFORMATEX structure based on maximum format size
// according to the ACM.
function TRecorderForm.AllocWaveFormatEx : Integer;
begin
    // get the largest format size required from installed ACMs...
    // FMaxFmtSize is the sum of sizeof(WAVEFORMATEX) + FWaveFormat.cbSize
    if acmMetrics(nil, ACM_METRIC_MAX_SIZE_FORMAT, FMaxFmtSize) <> 0 then begin
        ierrormsg('Error getting the max compression format size.');
        Result := -1;
        Exit;
    end;

    GetMem(FWaveFormat, FMaxFmtSize);
    if FWaveFormat = nil then begin
        ierrormsg('Error allocating local memory for WaveFormatEx structure.');
        Result := -2;
        Exit;
    end;

    // initialize the format to standard PCM...
    FillChar(FWaveFormat^, FMaxFmtSize, 0);
    FWaveFormat.wFormatTag      := WAVE_FORMAT_PCM;
    FWaveFormat.nChannels       := 1;
    FWaveFormat.nSamplesPerSec  := 11025;
    FWaveFormat.nAvgBytesPerSec := 11025;
    FWaveFormat.nBlockAlign     := 1;
    FWaveFormat.wBitsPerSample  := 8;
    FWaveFormat.cbSize          := 0;

    // store the format and tag decription strings...
    GetFormatTagDetails(FWaveFormat.wFormatTag);
    GetFormatDetails(FWaveFormat);

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Free the WAVEFORMATEX buffer.
procedure TRecorderForm.FreeWaveFormatEx;
begin
    if FWaveFormat <> nil then begin
        FreeMem(FWaveFormat);
        FWaveFormat := nil;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Allocate and lock header memory
function TRecorderForm.AllocWaveHeader : integer;
var
    i : Integer;
begin
    for i := Low(FWaveHdr) to High(FWaveHdr) do begin
        GetMem(FWaveHdr[i], sizeof(TWAVEHDR));
        if FWaveHdr[i] = nil then begin
            ierrormsg('Error allocating wave header memory.');
            Result := -1;
            Exit;
        end;
    end;
    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Free the wave header memory.
procedure TRecorderForm.FreeWaveHeader;
var
    i : Integer;
begin
    for i := Low(FWaveHdr) to High(FWaveHdr) do begin
        if FWaveHdr[i] <> nil then begin
            FreeMem(FWaveHdr[i]);
            FWaveHdr[i] := nil;
        end;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Allocate and lock wave memory
function TRecorderForm.AllocPCMBuffers : Integer;
var
    i : Integer;
begin
    for i := Low(FWaveMem) to High(FWaveMem) do begin
        GetMem(FWaveMem[i], WAVE_BUFSIZE);
        if FWaveMem[i] = nil then begin
            errormsg('Error allocating wave buffer memory.');
            Result := -1;
            Exit;
        end;
        FWaveHdr[i].lpData := FWaveMem[i];
    end;
    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Free the wave memory.
procedure TRecorderForm.FreePCMBuffers;
var
    i : Integer;
begin
    for i := Low(FWaveMem) to High(FWaveMem) do begin
        if FWaveMem[i] <> nil then begin
            FreeMem(FWaveMem[i]);
            FWaveMem[i] := nil;
        end;
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Write the recorded data to the temporary file.
function TRecorderForm.WriteWaveBuffer(size : UINT) : integer;
begin
    Result := 0;
    if size = 0 then
        Exit;

    if _lwrite(FTmpFileHandle, FWaveMem[FBufIndex], size) <> size then begin
        ierrormsg('Error writing data to temporary file.');
        Result := -1;
        Exit;
    end;

    // running total bytes recorded...
    Inc(FByteDataSize, size);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Create the temporary file for writing the wave data to.
function TRecorderForm.CreateTmpFile : integer;
var
    RootPathName          : array [0..MAX_PATH] of char;
    SectorsPerCluster     : DWORD;
    BytesPerSector        : DWORD;
    NumberOfFreeClusters  : DWORD;
    TotalNumberOfClusters : DWORD;
begin
    FTmpFileHandle := _lcreat(PChar(FTmpFileName), 0);
    if FTmpFileHandle = HFILE_ERROR then begin
        ierrormsg('Error creating temporary file.');
        Result := -1;
        Exit;
    end;

    // get available space on the temp disk...
    if FTmpFileName[2] = ':' then
        RootPathName[0] := FTmpFileName[1]
    else
        GetCurrentDirectory(sizeof(RootPathName), @RootPathName);
    RootPathName[1] := ':';
    RootPathName[2] := '\';
    RootPathName[3] := #0;

    GetDiskFreeSpace(@RootPathName,
                     SectorsPerCluster,
                     BytesPerSector,
                     NumberOfFreeClusters,
                     TotalNumberOfClusters);
    // We may have more than 2GB (High(Integer)) free space !
    if NumberOfFreeClusters > (High(DWORD) div SectorsPerCluster) then
        FDiskFreeSpace := High(DWORD)
    else begin
        FDiskFreeSpace := NumberOfFreeClusters * SectorsPerCluster;
        if FDiskFreeSpace > (High(DWORD) div BytesPerSector) then
            FDiskFreeSpace := High(DWORD)
        else
            FDiskFreeSpace := FDiskFreeSpace * BytesPerSector;
    end;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Close the temporary file contaning the newly recorded data.
procedure TRecorderForm.CloseTmpFile;
begin
    if _lclose(FTmpFileHandle) = HFILE_ERROR then
        errormsg('Error closing temporary file.');
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Delete the temporary wave file.
procedure TRecorderForm.DeleteTmpFile;
begin
    if Length(FTmpFileName) > 0 then
        DeleteFile(FTmpFileName);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Prepare headers, add buffer, adjust display, and start recording.
function TRecorderForm.StartWaveRecord : Integer;
var
    Status : MMRESULT;
begin
    FTotalWaveSize := 0;
    FByteDataSize  := 0;
    FBufIndex      := 0;

    // open the device for recording...
    Status := waveInOpen(@FWaveIn, WAVE_MAPPER, FWaveFormat,
                         Handle, 0, CALLBACK_WINDOW);
    if Status <> MMSYSERR_NOERROR then begin
        ierrormsg('Could not open the input device for recording.');
        Result := -1;
        Exit;
    end;

    // tell CloseWaveDeviceRecord() that the device is open...
    FDeviceOpened := TRUE;

    // prepare the headers...
    InitWaveHeaders;

    if not ((waveInPrepareHeader(FWaveIn, FWaveHdr[0], sizeof(TWAVEHDR)) = 0) and
        (waveInPrepareHeader(FWaveIn, FWaveHdr[1], sizeof(TWAVEHDR)) = 0))
    then begin
        CloseWaveDeviceRecord;
        ierrormsg('Error preparing header for recording.');
        Result := -2;
        Exit;
    end;

    // add the first buffer...
    if AddNextBuffer <> 0 then begin
        Result := -3;
        Exit;
    end;

    // create the file we'll be writing to...
    if CreateTmpFile <> 0 then begin
        CloseWaveDeviceRecord;
        Result := -4;
        Exit;
    end;

    // start recording to first buffer...
    if waveInStart(FWaveIn) <> 0 then begin
        CloseWaveDeviceRecord;
        ierrormsg('Error starting wave record.');
        Result := -5;
        Exit;
    end;

    FRecorderMode := recModeRecord;

    // set the timer for updating the display...
    Timer1.Interval := 250;
    Timer1.Enabled  := TRUE;
    UpdateLength(0, FDiskFreeSpace);

    // queue the next buffer...
    if AddNextBuffer <> 0 then begin
        Result := -6;
        Exit;
    end;

    // turn the record light on...
    RecordLight(recModeRecord);

    // adjust control button states...
    RecordButton.Caption := '&Stop';
    SetButtonState;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.SetButtonState;
begin
    RecordButton.Enabled := (FRecorderMode <> recModePlay);
    PlayButton.Enabled   := ((FTotalWaveSize <> 0) and
                             (FRecorderMode <> recModeRecord));
    AudioFormatButton.Enabled := (FRecorderMode = recModeOff);
//    EnableMenuItem( GetSubMenu(GetMenu(hwnd),0), 0, MF_BYPOSITION | ((FRecorderMode)?MF_DISABLED:MF_ENABLED) );
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Zero out the wave headers and initialize the data pointers and
// buffer lengths.
procedure TRecorderForm.InitWaveHeaders;
begin
    // make the wave buffer size a multiple of the block align...
    FWaveBufSize := (WAVE_BUFSIZE - WAVE_BUFSIZE mod FWaveFormat.nBlockAlign);

    // zero out the wave headers...
    FillChar(FWaveHdr[0]^, sizeof(TWAVEHDR), 0);
    FillChar(FWaveHdr[1]^, sizeof(TWAVEHDR), 0);

    // now init the data pointers and buffer lengths...
    FWaveHdr[0].dwBufferLength := FWaveBufSize;
    FWaveHdr[1].dwBufferLength := FWaveBufSize;
    FWaveHdr[0].lpData         := FWaveMem[0];
    FWaveHdr[1].lpData         := FWaveMem[1];
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Close the temp file and wave recording device.
procedure TRecorderForm.CloseWaveDeviceRecord;
begin
    // if the device is already closed, just return...
    if not FDeviceOpened then
        Exit;

    // unprepare the headers...
    if waveInUnprepareHeader(FWaveIn, FWaveHdr[0], sizeof(TWAVEHDR)) <> 0 then
        errormsg('Error in waveInUnprepareHeader (1)');

    if waveInUnprepareHeader(FWaveIn, FWaveHdr[1], sizeof(TWAVEHDR)) <> 0 then
        errormsg('Error in waveInUnprepareHeader (2)');

    // save the total size recorded and update the display...
    FTotalWaveSize := FByteDataSize;
    UpdateLength(FTotalWaveSize, FDiskFreeSpace);

    // tell the file save functions that we've got unsaved data...
    FRecordedData := TRUE;

    // close the temporary file...
    CloseTmpFile;

    // close the wave input device...
    if waveInClose(FWaveIn) <> 0 then
        errormsg('Error closing input device.');

    // tell this function we are now closed...
    FDeviceOpened := FALSE;

    // update display...
    // InvalidateRect( hwnd, &specrect, TRUE );
    RecordLight(recModeOff);
    RecordButton.Caption := '&Record';
    SetButtonState;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Add the buffer to the wave input queue and toggle buffer index. This
// routine is called from the main window proc.
function TRecorderForm.AddNextBuffer : integer;
begin
    // queue the buffer for input...
    if waveInAddBuffer(FWaveIn, FWaveHdr[FBufIndex], sizeof(TWAVEHDR)) <> 0 then begin
        StopWaveRecord;
        ierrormsg('Error adding buffer.');
        Result := -1;
        Exit;
    end;

    // toggle for next buffer...
    FBufIndex := 1 - FBufIndex;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Get the total bytes recorded and update the display with the information.
procedure TRecorderForm.UpdateRecordDisplay;
var
    mmtime : TMMTIME;
begin
    mmtime.wType := TIME_BYTES;
    waveInGetPosition(FWaveIn, @mmtime, sizeof(mmtime));
    UpdateLength(mmtime.cb, FDiskFreeSpace);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Get the total bytes played and update the display with the information.
procedure TRecorderForm.UpdatePlayDisplay;
var
    mmtime : TMMTIME;
begin
    mmtime.wType := TIME_BYTES;
    waveOutGetPosition(FWaveOut, @mmtime, sizeof(mmtime));
    UpdateLength(mmtime.cb, FTotalWaveSize);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.RecordLight(Mode : TRecorderMode);
const
    RecordColor : array [TRecorderMode] of TColor = (clBlack, clRed, clLime);
begin
    RecordLightShape.Brush.Color := RecordColor[Mode];
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Stop the recording.
procedure TRecorderForm.StopWaveRecord;
begin
    // set flag before stopping since it's used in the MM_WIM_DATA message
    // in our main window proc to control whether we add another buffer
    // or to close the device.
    FRecorderMode := recModeOff;

    // stop recording and return queued buffers...
    if waveInReset(FWaveIn) <> 0 then
        errormsg('Error in waveInReset');

    // stop the timer...
    Timer1.Enabled := FALSE;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.Timer1Timer(Sender: TObject);
begin
    if FRecorderMode = recModeRecord then
        UpdateRecordDisplay
    else if FRecorderMode = recModePlay then
        UpdatePlayDisplay;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.RecordButtonClick(Sender: TObject);
begin
    if FRecorderMode <> recModeOff then  // stop recording...
        StopWaveRecord
    else	                    // start recording...
        StartWaveRecord;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// done recording buffer...
procedure TRecorderForm.MMWimData(var msg: TMessage);
var
    pwavehdrtmp : PWAVEHDR;
begin
    // Done recording buffer, write it out
    pwavehdrtmp := PWAVEHDR(msg.lparam);
    if WriteWaveBuffer(pwavehdrtmp.dwBytesRecorded) <> 0 then
        StopWaveRecord;

    if FRecorderMode <> recModeOff then
        AddNextBuffer    	        // queue it again...
    else
        CloseWaveDeviceRecord;          // stop recording...
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// done playing queued wave buffer...
procedure TRecorderForm.MMWomDone(var msg: TMessage);
begin
    // done playing queued wave buffer...
    if FMoreToPlay then
        QueueNextBuffer
    else
        StopWavePlay(FALSE);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// closing wave output device...
procedure TRecorderForm.MMWomClose(var msg: TMessage);
begin
    FDeviceOpened := FALSE;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
    CheckSaveFile;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// If unsaved wave data has been recorded, ask the user to save it.
procedure TRecorderForm.CheckSaveFile;
begin
    if FRecordedData and (FTotalWaveSize <> 0) and (not FFileSaved) then begin
        if Application.MessageBox(
                            'Wave data has not been saved. ' +
                            'Do you want to save it?',
                            'Just checking', MB_YESNO) = IDYES then
            SaveWaveFile(smSaveAs);
    end;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Close the wave file, display an error message
procedure TRecorderForm.WFerror(
    mmfp      : HMMIO;
    const msg : String);
begin
    mmioClose(mmfp, 0);
    Cursor := crDefault;
    ierrormsg(msg);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Save the wave file by creating the required RIFF chunks and copy
// the wave data from the temporary wave file to the RIFF .wave file.
function TRecorderForm.SaveWaveFile(FSaveMode : TRecorderSaveMode) : Integer;
var
    mmfp           : HMMIO;
    dwTotalSamples : DWORD;
    fTotalSamples  : double;
    mminfopar      : TMMCKINFO;
    mminfosub      : TMMCKINFO;
const
    initsavename : Boolean = FALSE;
begin
    // if no data recorded, don't bother...
    if FTotalWaveSize = 0 then begin
        ierrormsg('No recorded wave data to save.');
        Result := 0;
        Exit;
    end;

    // initialize the save file struct...
    if not initsavename then begin
        InitFilenameDialog(OpenDialog1);
        initsavename := TRUE;
    end;

    // get a filename/use the current one, or cancel...
    if GetSaveFilename(FSaveMode) = FALSE then begin
        Result := 0;
        Exit;
    end;

    // open the wave file for write...
    mmfp := mmioOpen(PChar(FFilename), nil,
                     MMIO_CREATE or MMIO_WRITE or MMIO_ALLOCBUF);
    if mmfp = 0 then begin
        ierrormsg('Error opening file for write.');
        Result := -1;
        Exit;
    end;

    Cursor := crHourGlass;

    // create wave RIFF chunk...
    mminfopar.fccType := FOURCC_WAVE;
    mminfopar.cksize := 0;		 	// let the function determine the size
    if mmioCreateChunk(mmfp, @mminfopar, MMIO_CREATERIFF) <> 0 then begin
        WFerror(mmfp, 'Error creating RIFF wave chunk.');
        Result := -2;
        Exit;
    end;

    // create the format chunk and write the wave format...
    mminfosub.ckid   := FOURCC_FMT;
    mminfosub.cksize := FMaxFmtSize;
    if mmioCreateChunk(mmfp, @mminfosub, 0) <> 0 then begin
        WFerror(mmfp, 'Error creating RIFF format chunk.');
        Result := -3;
        Exit;
    end;

    if mmioWrite(mmfp, PChar(FWaveFormat), FMaxFmtSize) <> LongInt(FMaxFmtSize) then begin
        WFerror(mmfp, 'Error writing RIFF format data.');
        Result := -3;
        Exit;
    end;

    // back out of format chunk...
    mmioAscend(mmfp, @mminfosub, 0);

    // write the fact chunk (required for all non-PCM .wav files...
    // this chunk just contains the total length in samples...
    mminfosub.ckid   := FOURCC_FACT;
    mminfosub.cksize := sizeof(DWORD);
    if mmioCreateChunk(mmfp, @mminfosub, 0) <> 0 then begin
        WFerror(mmfp, 'Error creating RIFF ''fact'' chunk.');
        Result := -4;
        Exit;
    end;

    fTotalSamples := FTotalWaveSize / FWaveFormat.nAvgBytesPerSec *
                     FWaveFormat.nSamplesPerSec;
    dwTotalSamples := Trunc(fTotalSamples);
    if mmioWrite(mmfp, PChar(@dwTotalSamples), sizeof(dwTotalSamples))
       <> sizeof(dwTotalSamples) then begin
        WFerror(mmfp, 'Error writing RIFF ''fact'' data.');
        Result := -4;
        Exit;
    end;

    // back out of fact chunk...
    mmioAscend(mmfp, @mminfosub, 0);

    // now create and write the wave data chunk...
    mminfosub.ckid   := FOURCC_DATA;
    mminfosub.cksize := 0;	 	// let the function determine the size
    if mmioCreateChunk(mmfp, @mminfosub, 0) <> 0 then begin
        WFerror(mmfp, 'Error creating RIFF data chunk.');
        Result := -5;
        Exit;
    end;

    // copy the data from the temp file to the wave file...
    if CopyDataToWaveFile(mmfp) <> 0 then begin
        WFerror(mmfp, 'Error writing wave data.');
        Result := -5;
        Exit;
    end;

    // back out and cause the size of the data buffer to be written...
    mmioAscend(mmfp, @mminfosub, 0);

    // ascend out of the RIFF chunk...
    mmioAscend(mmfp, @mminfopar, 0);

    // done...
    mmioClose(mmfp, 0);

    Cursor := crDefault;

    // indicate filename on window title...
    Caption := WindowCaption + ' - ' + FFileName;

    // set this so we don't ask for a filename when the user does "SAVE"...
    FFileSaved := TRUE;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Initialize non-variant fields of OPENFILENAME struct.
procedure TRecorderForm.InitFilenameDialog(OpenDialog : TOpenDialog);
begin
    // just do this once...
    OpenDialog.Filter      := 'Wave Files (*.wav)|*.wav|All Files (*.*)|*.*';
    OpenDialog.FileName    := FFilename;
    OpenDialog.InitialDir  := ExtractFilePath(FFilename);
    OpenDialog.Title       := WindowCaption;
    OpenDialog.DefaultExt  := 'WAV';
    OpenDialog.Options     := [ofHideReadOnly];
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Get a save filename according to the save mode.  The FSaveMode is either
// "save" or "saveas" where "saveas" queries for a new name and "save" checks
// to see if the file exists and checks with the user before saving.
// Returns:
//	TRUE	- We have a filename, use that for the save.
//	FALSE	- Cancel the save operation.
function TRecorderForm.GetSaveFilename(FSaveMode : TRecorderSaveMode) : Boolean;
var
    response : Integer;
begin
    while TRUE do begin
        // if we're saving the file as the current name...
        if FSaveMode = smSave then begin
            if (not FHasFilename) or (Length(FFilename) = 0) then
                FSaveMode := smSaveAs
            else if not FFileSaved then begin // see if there's a file by that name already...
                if FileExists(FFilename) then begin	// there is. overwrite?
                    response := QueryUserSave(FFilename);
                    if response = IDNO then
                        FSaveMode := smSaveAs		// no, get a filename
                    else begin
                        Result := (response = IDYES);	// cancel the save operation
                        Exit;
                    end;
                end;
            end;
        end;

        if FSaveMode = smSaveAs then begin		// get a save filename...
            InitFilenameDialog(OpenDialog1);                // setup for common dlg
            if not OpenDialog1.Execute then begin
                Result := FALSE;
                Exit;
            end;

            FFilename    := OpenDialog1.FileName;
            FHasFilename := TRUE;

            // now go check to see if the files exists...
            FSaveMode := smSave;
            continue;
        end;

        break;
    end;

    Result := TRUE;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Notify user that the file exists and ask to overwrite, return 1 for yes, 0 for no.
function TRecorderForm.QueryUserSave(const name : String) : Integer;
begin
    Result := Application.MessageBox(
                  PChar('File ' + name + ' exists, overwite?'),
                  'Save File', MB_YESNOCANCEL);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Copy the wave data in the temporary file to the wave file.
function TRecorderForm.CopyDataToWaveFile(mmfp : HMMIO) : integer;
var
    pbuf   : PChar;
    ht     : HFILE;
    nbytes : integer;
begin
    pbuf := FWaveMem[0]; // use one of the wave buffers for copying
    // open the temp file for reading...
    ht := _lopen(PChar(FTmpFileName), OF_READ);
    if ht = HFILE_ERROR then begin
        Result := -1;
        Exit;
    end;

    // copy to RIFF/wave file...
    while TRUE do begin
        nbytes := _lread(ht, pbuf, WAVE_BUFSIZE);
        if nbytes <= 0 then
            break;
        mmioWrite(mmfp, pbuf, nbytes);
    end;

    // close read file...
    _lclose(ht);

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.Save1Click(Sender: TObject);
begin
    SaveWaveFile(smSave);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.SaveAs1Click(Sender: TObject);
begin
    SaveWaveFile(smSaveAs);
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.Exit1Click(Sender: TObject);
begin
    Close;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Prepare headers, add buffer, and start recording.
function TRecorderForm.StartWavePlay : Integer;
begin
    Result := -1;

    // if the device is still open, just return...
    if FDeviceOpened then
        Exit;

    FByteDataSize := 0;
    FBufIndex     := 0;

    // open the device for recording...
    if waveOutOpen(@FWaveOut, WAVE_MAPPER, FWaveFormat,
                   Handle, 0, CALLBACK_WINDOW or WAVE_ALLOWSYNC) <> 0 then begin
        ierrormsg('Error opening wave out device.');
        Result := -1;
        Exit;
    end;

    FDeviceOpened := TRUE;

    // prepare the headers...
    InitWaveHeaders;

    if (waveOutPrepareHeader(FWaveOut, FWaveHdr[0], sizeof(TWAVEHDR)) <> 0) or
       (waveOutPrepareHeader(FWaveOut, FWaveHdr[1], sizeof(TWAVEHDR)) <> 0)
    then begin
        CloseWaveDevicePlay;
        ierrormsg('Error preparing header for playing.');
        Result := -2;
        Exit;
    end;

    // open the temporary file for reading...
    if OpenTmpFile = 0 then begin
        CloseWaveDevicePlay;
        ierrormsg('Error opening temporary wave file for read.');
        Result := -3;
        Exit;
    end;

    // write the first buffer to start playing,..
    if QueueNextBuffer <> 0 then begin
        CloseWaveDevicePlay;
        Result := -4;
        Exit;
    end;

    FRecorderMode := recModePlay;
    FMoreToPlay   := TRUE;

    // set the timer for updating the display...
    Timer1.Interval := 250;
    Timer1.Enabled  := TRUE;
    UpdateLength(0, FTotalWaveSize);

    // turn the play light on...
    RecordLight(recModePlay);

    // adjust control button states...
    PlayButton.Caption := '&Stop';
    SetButtonState;

    // and queue the next buffer up...
    QueueNextBuffer;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Write the buffer to the wave device and toggle buffer index.
function TRecorderForm.QueueNextBuffer : Integer;
begin
    // fill the wave buffer with data from the file...
    if ReadWaveBuffer = 0 then begin
        // reset flags field (remove WHDR_DONE attribute)...
        FWaveHdr[FBufIndex].dwFlags := WHDR_PREPARED;

        // now queue the buffer for output...
        if waveOutWrite(FWaveOut, FWaveHdr[FBufIndex], sizeof(TWAVEHDR)) <> 0
        then begin
            StopWavePlay(TRUE);
            ierrormsg('Error writing wave buffer.');
            Result := -1;
            Exit;
        end;

        // toggle for next buffer...
        FBufIndex := 1 - FBufIndex;
    end;

    Result := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Read a wave chunk from the temp file.
function TRecorderForm.ReadWaveBuffer : Integer;
begin
    // if we haven't encountered the end of the wave yet,
    // read another buffer in...
    if FByteDataSize < FTotalWaveSize then begin
        // read wave chunk from the temporary file...
        FWaveHdr[FBufIndex].dwBufferLength :=
            _lread(FTmpFileHandle, FWaveMem[FBufIndex], FWaveBufSize);

        // update total number of bytes read so far...
        Inc(FByteDataSize, FWaveHdr[FBufIndex].dwBufferLength);
        Result := 0;
        Exit;
    end;

    // otherwise the last buffer has been queued, just let it finish playing...
    FMoreToPlay := FALSE;	// handled in MM_WOM_DONE message
    Result      := 1;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Open the temporary file for reading.
function TRecorderForm.OpenTmpFile : Integer;
begin
    FTmpFileHandle := _lopen(PChar(FTmpFileName), OF_READ);
    if FTmpFileHandle = HFILE_ERROR then
        Result := 0
    else
        Result := 1;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Stop the wave playing.
procedure TRecorderForm.StopWavePlay(prematurely : Boolean);
begin
    // if the device isn't open, just return...
    if not FDeviceOpened then
        Exit;

    // stop playing...
    waveOutReset(FWaveOut);

    // stop the timer...
    Timer1.Enabled := FALSE;

    // update the length only if we played to the end...
    if not prematurely then
        UpdateLength(FByteDataSize, FTotalWaveSize);

    FRecorderMode := recModeOff;
    FMoreToPlay   := FALSE;

    // close the device and unprepare the headers...
    CloseWaveDevicePlay;

    // update display...
    RecordLight(recModeOff);
    PlayButton.Caption := '&Play';
    SetButtonState;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Close the wave output device.
procedure TRecorderForm.CloseWaveDevicePlay;
begin
    // unprepare the headers...
    if (waveOutUnprepareHeader(FWaveOut, FWaveHdr[0], sizeof(TWAVEHDR)) <> 0) or
       (waveOutUnprepareHeader(FWaveOut, FWaveHdr[1], sizeof(TWAVEHDR)) <> 0)
    then
        errormsg('Error unpreparing play header.');

    // close the device...
    if waveOutClose(FWaveOut) <> 0 then
        errormsg('Error closing wave play device.');

    FDeviceOpened := FALSE;

    // close the temporary file...
    CloseTmpFile;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.PlayButtonClick(Sender: TObject);
begin
    if FRecorderMode = recModePlay then // stop playing...
        StopWavePlay(TRUE)
    else		           // start playing...
        StartWavePlay;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.Open1Click(Sender: TObject);
begin
    ReadWaveFile;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
function TRecorderForm.ReadWaveFile : Integer;
var
    mmfp           : HMMIO;
    mminfopar      : TMMCKINFO;
    mminfosub      : TMMCKINFO;
begin
    InitFilenameDialog(OpenDialog1);
    if not OpenDialog1.Execute then begin
        Result := 0;
        Exit;
    end;
    FFileName    := OpenDialog1.FileName;
    FHasFilename := TRUE;

    // open the wave file for reading...
    mmfp := mmioOpen(PChar(FFileName), nil, MMIO_READ + MMIO_ALLOCBUF);
    if mmfp = 0 then begin
        ierrormsg('Open for read failed.');
        Result := -1;
        Exit;
    end;

    Cursor := crHourGlass;

    // search for wave type and format chunks...
    mminfopar.fccType := FOURCC_WAVE;
    if mmioDescend(mmfp, @mminfopar, nil, MMIO_FINDRIFF) <> 0 then begin
        WFerror(mmfp, 'Wave format not found in file.');
        Result := -2;
        Exit;
    end;

    mminfosub.ckid := FOURCC_FMT;
    if mmioDescend(mmfp, @mminfosub, @mminfopar, MMIO_FINDCHUNK) <> 0 then begin
        WFerror(mmfp, 'Format chunk not found.');
        Result := -3;
        Exit;
    end;

    // read the wave format if the size checks out...
    if mminfosub.cksize > FMaxFmtSize then begin
        WFerror(mmfp, 'Format size in file doesn''t match ACM''s.');
        Result := -4;
        Exit;
    end;

    // read the wave format...
    if mmioRead(mmfp, PChar(FWaveFormat), mminfosub.cksize)
          <> LongInt(mminfosub.cksize) then begin
        WFerror(mmfp, 'Error reading format chunk.');
        Result := -4;
        Exit;
    end;

    // get the total wave data size (mminfo.cksize)...
    mmioAscend(mmfp, @mminfosub, 0);
    mminfosub.ckid := FOURCC_DATA;
    if mmioDescend(mmfp, @mminfosub, @mminfopar, MMIO_FINDCHUNK) <> 0 then begin
        WFerror(mmfp, 'Data chunk not found.');
        Result := -5;
        Exit;
    end;

    // if there's no data get out...
    if mminfosub.cksize = 0 then begin
        WFerror(mmfp, 'The data chunk contains no data.');
        Result := -6;
        Exit;
    end;

    // now read the wave data and copy it to the temorary file...
    if CopyWaveToTempFile(mmfp, mminfosub.cksize) <> 0 then begin
        mminfosub.cksize := 0;
        errormsg('Error reading the wave data.');
    end;

    // all done...
    mmioClose(mmfp, 0);
    Cursor := crDefault;
    FTotalWaveSize := mminfosub.cksize;

    // this data wasn't recorded, it was read in...
    FRecordedData := FALSE;

    // store the format and tag decription strings...
    GetFormatTagDetails(FWaveFormat.wFormatTag);
    GetFormatDetails(FWaveFormat);

    // update display...
    UpdateLength(FTotalWaveSize, FTotalWaveSize);
    SetButtonState;
    FormatTagLabel.Caption  := FFormatTag;
    FormatDescLabel.Caption := FFormatDesc;

    Caption := WindowCaption + ' - ' + FFileName;
    Result  := 0;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
// Copy the wave data from the RIFF file to the temorary file.
function TRecorderForm.CopyWaveToTempFile(
    mmfp     : HMMIO;
    datasize : DWORD) : Integer;
var
    pbuf     : PChar;
    ntotal   : DWORD;
    nbytes   : integer;
    readsize : DWORD;
begin
    pbuf     := FWaveMem[0];
    readsize := WAVE_BUFSIZE;
    ntotal   := 0;
    Result   := 0;

    // create the temporary file based on the current temp name...
    if CreateTmpFile <> 0 then begin
        Result := -1;
        Exit;
    end;

    // enter read/copy loop...
    while ntotal < datasize do begin
        if (ntotal + readsize) > datasize then
            readsize := datasize - ntotal;

        nbytes := mmioRead(mmfp, pbuf, readsize);
        if nbytes = 0 then begin
            Result := -2;
            break;
        end;

        if _lwrite(FTmpFileHandle, pbuf, nbytes) <> UINT(nbytes) then begin
            Result := -3;
            break;
        end;

        Inc(ntotal, nbytes);
    end;

    // close read file...
    CloseTmpFile;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
procedure TRecorderForm.Help1Click(Sender: TObject);
begin
    AboutBoxForm.ShowModal;
end;


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}

end.

